/*	FUZZauth - http://unfix.org/projects/fuzzauth & http://fuzzauth.sourceforge.net
 * FUZZauth_common.c by Jeroen Massar <jeroen@unfix.org>
 *
 * Common functions for communicating with a FUZZauth server over a UNIX socket.
*/

#include "fuzzauth_common.h"
#include <syslog.h>
#include <signal.h>

/* Globals */
volatile SOCKET			g_xServSock = INVALID_SOCKET;
volatile int				g_nConnectCount = 0;

void							(*g_xOldSignal)(int);
struct faPacket_Request	g_xPkt_Request;
struct faPacket_Answer	g_xPkt_Answer;
struct faPacket_Number	g_xPkt_Number;
struct faPacket_Number2	g_xPkt_Number2;
struct faPacket_String	g_xPkt_String;
struct faPacket_User		g_xPkt_User;
struct faPacket_Group	g_xPkt_Group;

/************************
 Logging Function
************************/
void fuzzauth_LogA(int nPriority, const char *sFormat, va_list xArgs)
{
	openlog(FUZZAUTH_NAME, LOG_PID, LOG_AUTH );
	vsyslog(nPriority, sFormat, xArgs);
	closelog();
}

void fuzzauth_Log(int nPriority, const char *sFormat, ... )
{
	va_list xArgs;
	va_start(xArgs, sFormat);
	fuzzauth_LogA(nPriority, sFormat, xArgs);
	va_end(xArgs);
}

#ifdef DEBUG
/************************
 Debug Functions
************************/
void DLOG(const char *sFormat, ... )
{
	va_list xArgs;
	va_start(xArgs, sFormat);
	fuzzauth_LogA(LOG_DEBUG, sFormat, xArgs);
	va_end(xArgs);
}
#endif

void FA_sigpipe(int i)
{
	D(DLOG("FA_sigpipe() - I just recieved a sigpipe (%d)...\n", i));
	signal(SIGPIPE, &FA_sigpipe);
	return;
	i = i;
}

/************************
 Interface Functions
************************/
bool fuzzauth_Connect(void)
{
	return fuzzauth_ConnectA((SOCKET *)&g_xServSock, (int *)&g_nConnectCount);
}

bool fuzzauth_ConnectA(SOCKET *sSocket, int *nConnectCount)
{
	bool				bRet = false;
	struct passwd	*xPassword;
	struct group	*xGroup;

//	D(DLOG("fuzzauth_Connect()\n"));

	if (*nConnectCount == 0)
	{
		// Determine the UID of the user who should own the socket.
		if ((xPassword = getpwnam(FUZZAUTH_NAME_LOW)))
		{
			// Determine the GID of the group we should be.
			if ((xGroup = getgrnam(FUZZAUTH_NAME_LOW)))
			{
				struct stat xStat;
				if (lstat(FUZZAUTH_UNIXSOCKET, &xStat) == 0)
				{
					// Is it really a socket?
					if (S_ISSOCK(xStat.st_mode))
					{
						// Is it owned by 'fuzzauth:fuzzauth' ?
						if ((xStat.st_uid == xPassword->pw_uid) && (xStat.st_gid == xGroup->gr_gid))
						{
							// Set us up with the UNIX socket.
							*sSocket = socket(PF_UNIX, SOCK_STREAM, 0);
							if (*sSocket != INVALID_SOCKET)
							{
								struct sockaddr_un xSockAddrUnix;
								memset(&xSockAddrUnix, 0, sizeof(xSockAddrUnix));
								xSockAddrUnix.sun_family = AF_UNIX;
								strncpy(xSockAddrUnix.sun_path, FUZZAUTH_UNIXSOCKET, sizeof(xSockAddrUnix.sun_path) - 1);

								if (connect(*sSocket, (struct sockaddr *)&xSockAddrUnix, sizeof(xSockAddrUnix)) != SOCKET_ERROR)
								{
									bRet = true;
									*nConnectCount = 1;
									D(DLOG("fuzzauth_Connect() - Connected, connect count now %d, Socket=%x\n", *nConnectCount, *sSocket));
									g_xOldSignal = signal(SIGPIPE, &FA_sigpipe);
								}
								else
								{
									closesocket(*sSocket);
									*sSocket = INVALID_SOCKET;
									*nConnectCount = 0;
									fuzzauth_Log(LOG_ERR, "Couldn't connect to daemon\n");
								}
							}
							else fuzzauth_Log(LOG_ERR, "Couldn't get an AF_UNIX socket\n");
					//		D(DLOG("fuzzauth_Connect() - %s\n", bRet == true ? "Success" : "Failure"));
						}
						else fuzzauth_Log(LOG_ERR, "The owner of the socket \"%s\" isn't correct, it should be %d:%d (%s:%s) and is %d:%d.\n", FUZZAUTH_UNIXSOCKET, xPassword->pw_uid, xGroup->gr_gid, FUZZAUTH_NAME_LOW, FUZZAUTH_NAME_LOW, xStat.st_uid, xStat.st_gid);
					}
					else fuzzauth_Log(LOG_ERR, "The socket \"%s\" isn't really a socket.\n", FUZZAUTH_UNIXSOCKET);
				}
				else fuzzauth_Log(LOG_ERR, "Couldn't stat() \"%s\", this probably means that the FUZZauth Daemon is not running.\n", FUZZAUTH_UNIXSOCKET);
			}
			else fuzzauth_Log(LOG_ERR, "Group \"%s\" doesn't exist at all.... read the doc/INSTALL file again!\n", FUZZAUTH_NAME_LOW);
		}
		else fuzzauth_Log(LOG_ERR, "User \"%s\" doesn't exist at all.... read the doc/INSTALL file again!\n", FUZZAUTH_NAME_LOW);
	}
	else
	{
		bRet = true;
		*nConnectCount=*nConnectCount+1;
		D(DLOG("fuzzauth_Connect() - Extra connect detected, connect count now %d, Socket=%x\n", *nConnectCount, *sSocket));
	}
	return bRet;
}

// Used for TCP/IP Connections
bool fuzzauth_ConnectB(SOCKET *sSocket, int *nConnectCount, char *sHost, char *sPort, int nFamily)
{
	bool bRet = false;

	D(DLOG("fuzzauth_ConnectB()\n"));
	if (*nConnectCount == 0)
	{
		struct addrinfo	xHints, *xAddrInfo = NULL, *xAI = NULL;
		int					nRetVal;

		// By setting the AI_PASSIVE flag in the hints to getaddrinfo, we're
		// indicating that we intend to use the resulting address(es) to bind
		// to a socket(s) for accepting incoming connections.  This means that
		// when the Address parameter is NULL, getaddrinfo will return one
		// entry per allowed protocol family containing the unspecified address
		// for that family.
		memset(&xHints, 0, sizeof(xHints));
		xHints.ai_family = nFamily;
		xHints.ai_socktype = SOCK_STREAM;
		nRetVal = getaddrinfo(sHost, sPort, &xHints, &xAddrInfo);
		if (nRetVal != 0) fuzzauth_Log(LOG_ERR, "fuzzauth_ConnectB() - getaddrinfo failed with error %d: %s\n", nRetVal, gai_strerror(nRetVal));
		else
		{
			// For each address getaddrinfo returned, we create a new socket,
			// bind that address to it, and create a queue to listen on.
			for (xAI = xAddrInfo; (*sSocket == SOCKET_ERROR) && (xAI != NULL); xAI = xAI->ai_next)
			{
				// We only support PF_INET and PF_INET6.
				if ((xAI->ai_family == PF_INET) || (xAI->ai_family == PF_INET6))
				{
					// Open a socket with the correct address family for this address.
					*sSocket = socket(xAI->ai_family, xAI->ai_socktype, xAI->ai_protocol);
					if (*sSocket != INVALID_SOCKET)
					{
						if (connect(*sSocket, xAI->ai_addr, xAI->ai_addrlen) != SOCKET_ERROR)
						{
							fuzzauth_Log(LOG_INFO, "fuzzauth_ConnectB() - Connected to host %s, port %s, protocol family %s\n", sHost, sPort, (xAI->ai_family == PF_INET) ? "PF_INET" : "PF_INET6");
							*nConnectCount = 1;
							bRet = true;
						}
						else
						{
							fuzzauth_Log(LOG_ERR, "fuzzauth_ConnectB() - connect() failed with %d: %s\n", ERRORNUMSTR);
							closesocket(*sSocket);
							*sSocket = INVALID_SOCKET;
							errno = 0;
						}
					}
					else fuzzauth_Log(LOG_ERR, "fuzzauth_ConnectB() - socket() failed with error %d: %s\n", ERRORNUMSTR);
				}
				D(else DLOG("fuzzauth_ConnectB() - Unknown address family %d\n", xAI->ai_family));
			}
			freeaddrinfo(xAddrInfo);
		}
	}
	else
	{
		D(DLOG("fuzzauth_ConnectB() - ConnectCount = %d\n", *nConnectCount));
		*nConnectCount = *nConnectCount+1;
	}
	return bRet;
}

void fuzzauth_InitPackets(int nCommand)
{
//	D(DLOG("fuzzauth_InitPackets(%d)\n", nCommand));
	memset(&g_xPkt_Request, 0, sizeof(struct faPacket_Request));
	memset(&g_xPkt_Answer, 0, sizeof(struct faPacket_Answer));
	g_xPkt_Request.m_nMagic[0] = 'F';
	g_xPkt_Request.m_nMagic[1] = 'U';
	g_xPkt_Request.m_nMagic[2] = 'Z';
	g_xPkt_Request.m_nMagic[3] = 'Z';
	g_xPkt_Request.m_nVersion = FUZZAUTH_VERSION;
	g_xPkt_Request.m_nCommand = nCommand;
//	D(DLOG("fuzzauth_InitPackets(%d) - Done\n", nCommand));
}

bool fuzzauth_GetAnswerA(SOCKET sSocket)
{
	bool bRet = false;
	D(DLOG("fuzzauth_GetAnswerA()\n"));
	if (read(sSocket, &g_xPkt_Answer, sizeof(struct faPacket_Answer)) != SOCKET_ERROR)
	{
		// Fixup endianness in the Answer packet.
		g_xPkt_Answer.m_nAnswer		= ntohl(g_xPkt_Answer.m_nAnswer);
		g_xPkt_Answer.m_nDatatype	= ntohl(g_xPkt_Answer.m_nDatatype);
		g_xPkt_Answer.m_nLength		= ntohl(g_xPkt_Answer.m_nLength);

		if (g_xPkt_Answer.m_nAnswer == c_nAnswer_Granted)
		{
			bRet = true;
		}
		D(else DLOG("fuzzauth_GetAnswerA() - Answer %d received\n", g_xPkt_Answer.m_nAnswer));
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_GetAnswerA() - Getting answer from daemon failed\n");
	D(DLOG("fuzzauth_GetAnswerA() = %s\n", bRet ? "Success" : "Failure"));
	return bRet;
}

bool fuzzauth_GetAnswerB(SOCKET sSocket, enum c_nDatatype nDatatype)
{
	bool bRet = false;

	D(DLOG("fuzzauth_GetAnswerB(%d)\n", nDatatype));
	if (g_xPkt_Answer.m_nDatatype == nDatatype)
	{
		void *xPkt = NULL;
		unsigned int nLength = 0;
		switch (nDatatype)
		{
			case c_nDatatype_Number:	xPkt = &g_xPkt_Number;
												nLength = sizeof(struct faPacket_Number);
												break;
			case c_nDatatype_Number2:	xPkt = &g_xPkt_Number2;
												nLength = sizeof(struct faPacket_Number2);
												break;
			case c_nDatatype_String:	xPkt = &g_xPkt_String;
												nLength = sizeof(struct faPacket_String);
												break;
			case c_nDatatype_User:		xPkt = &g_xPkt_User;
												nLength = sizeof(struct faPacket_User);
												break;
			case c_nDatatype_Group:		xPkt = &g_xPkt_Group;
												nLength = sizeof(struct faPacket_Group);
												break;
			case c_nDatatype_Unknown:	xPkt = NULL;
												nLength = 0;
												break;
		}
		if (g_xPkt_Answer.m_nLength == nLength)
		{
			if (read(sSocket, xPkt, nLength) != SOCKET_ERROR)
			{
				bRet = true;
				switch (nDatatype)
				{
					case c_nDatatype_Number:	
														// Fixup endianness in the Answer packet.
														g_xPkt_Number.m_nNumber = ntohl(g_xPkt_Number.m_nNumber);
														D(DLOG("fuzzauth_GetAnswer() - Number: %d\n", g_xPkt_Number.m_nNumber));
														break;
					case c_nDatatype_Number2:	// Fixup endianness in the Answer packet.
														g_xPkt_Number2.m_nNumber1 = ntohl(g_xPkt_Number2.m_nNumber1);
														g_xPkt_Number2.m_nNumber2 = ntohl(g_xPkt_Number2.m_nNumber2);
														D(DLOG("fuzzauth_GetAnswer() - Number2: %d, %d\n", g_xPkt_Number2.m_nNumber1, g_xPkt_Number2.m_nNumber2));
														break;
					case c_nDatatype_String:
														D(DLOG("fuzzauth_GetAnswer() - String: %s\n", g_xPkt_String.m_sString));
														break;
					case c_nDatatype_User:	
														// Fixup endianness in the Answer packet.
														g_xPkt_User.m_nUID = ntohl(g_xPkt_User.m_nUID);
														g_xPkt_User.m_nGID = ntohl(g_xPkt_User.m_nGID);
														D(DLOG("fuzzauth_GetAnswer() - User: UID=%d, GID=%d, Name=%s\n", g_xPkt_User.m_nUID, g_xPkt_User.m_nGID, g_xPkt_User.m_sUsername));
														break;
					case c_nDatatype_Group:
														// Fixup endianness in the Answer packet.
														g_xPkt_Group.m_nGID = ntohl(g_xPkt_Group.m_nGID);
														g_xPkt_Group.m_nNumMembers = ntohl(g_xPkt_Group.m_nNumMembers);
														D(DLOG("fuzzauth_GetAnswer() - Group: GID=%d, Name=%s\n", g_xPkt_Group.m_nGID, g_xPkt_Group.m_sGroupname));
														break;
					case c_nDatatype_Unknown:
														D(DLOG("fuzzauth_GetAnswer() - Unknown: %d\n", nDatatype)); break;
				}
			}
			else fuzzauth_Log(LOG_ERR, "fuzzauth_GetAnswerB() - Getting data from daemon failed\n");
		}
		else fuzzauth_Log(LOG_ERR, "fuzzauth_GetAnswerB() - Data Length Mismatch\n");
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_GetAnswerB() - Unexpected datatype %d received, expected %d\n", g_xPkt_Answer.m_nDatatype, nDatatype);
	return bRet;
}

bool fuzzauth_Request(SOCKET sSocket, bool bGetAnswer)
{
	bool bRet = false;
	void *xPkt = NULL;

	D(DLOG("fuzzauth_Request()\n"));
	switch (g_xPkt_Request.m_nCommand)
	{
		case c_nCommand_Quit:
		case c_nCommand_GetUsers:
		case c_nCommand_GetGroups:	
													xPkt = NULL;
													g_xPkt_Request.m_nDatatype = c_nDatatype_Unknown;
													g_xPkt_Request.m_nLength = 0;
													break;
		case c_nCommand_Mode:
		case c_nCommand_MatchUserID:
		case c_nCommand_MatchGroupID:
													xPkt = &g_xPkt_Number;
													g_xPkt_Request.m_nDatatype = c_nDatatype_Number;
													g_xPkt_Request.m_nLength = sizeof(struct faPacket_Number);
													g_xPkt_Number.m_nNumber = htonl(g_xPkt_Number.m_nNumber);
													break;
		case c_nCommand_MemberAdd:
		case c_nCommand_MemberRemove:
													xPkt = &g_xPkt_Number2;
													g_xPkt_Request.m_nDatatype = c_nDatatype_Number2;
													g_xPkt_Request.m_nLength = sizeof(struct faPacket_Number2);
													g_xPkt_Number2.m_nNumber1 = htonl(g_xPkt_Number2.m_nNumber1);
													g_xPkt_Number2.m_nNumber2 = htonl(g_xPkt_Number2.m_nNumber2);
													break;
		case c_nCommand_MatchUserName:
		case c_nCommand_MatchUserGECOS:
		case c_nCommand_MatchGroupName:
		case c_nCommand_UserRemove:
		case c_nCommand_GroupRemove:
													xPkt = &g_xPkt_String;
													g_xPkt_Request.m_nDatatype = c_nDatatype_String;
													g_xPkt_Request.m_nLength = sizeof(struct faPacket_String);
													break;
		case c_nCommand_Authenticate:
		case c_nCommand_ChangePassword:
		case c_nCommand_UserAdd:
		case c_nCommand_UserChange:
													xPkt = &g_xPkt_User;
													g_xPkt_Request.m_nDatatype = c_nDatatype_User;
													g_xPkt_Request.m_nLength = sizeof(struct faPacket_User);
													g_xPkt_User.m_nUID = htonl(g_xPkt_User.m_nUID);
													g_xPkt_User.m_nGID = htonl(g_xPkt_User.m_nGID);
													break;
		case c_nCommand_GroupAdd:
		case c_nCommand_GroupChange:
													xPkt = &g_xPkt_Group;
													g_xPkt_Request.m_nDatatype = c_nDatatype_Group;
													g_xPkt_Request.m_nLength = sizeof(struct faPacket_Group);
													g_xPkt_Group.m_nGID = htonl(g_xPkt_Group.m_nGID);
													g_xPkt_Group.m_nNumMembers = htonl(g_xPkt_Group.m_nNumMembers);
													break;
		// Unused cases
		case c_nCommand_Unknown:			break;
	}
	if (g_xPkt_Request.m_nCommand != c_nCommand_Unknown)
	{
		// Fixup endianness in the Request packet.
		g_xPkt_Request.m_nVersion	= htonl(g_xPkt_Request.m_nVersion);
		g_xPkt_Request.m_nCommand	= htonl(g_xPkt_Request.m_nCommand);
		g_xPkt_Request.m_nDatatype	= htonl(g_xPkt_Request.m_nDatatype);
		g_xPkt_Request.m_nLength	= htonl(g_xPkt_Request.m_nLength);

		if (!(write(sSocket, &g_xPkt_Request, sizeof(struct faPacket_Request)) == SOCKET_ERROR))
		{
			D(DLOG("fuzzauth_Request() - Sending request to daemon succeeded\n"));
			if (g_xPkt_Request.m_nLength != 0)
			{
				if (write(sSocket, xPkt, ntohl(g_xPkt_Request.m_nLength)) != SOCKET_ERROR)
				{
					D(DLOG("fuzzauth_Request() - Sending data request to daemon succeeded\n"));
					bRet = true;
				}
				else fuzzauth_Log(LOG_ERR, "fuzzauth_Request() - Sending data request to daemon failed: %d: %s\n", ERRORNUMSTR);
			}
			else bRet = true;
			if ((bRet) && (bGetAnswer))
			{
				D(DLOG("fuzzauth_Request() - Waiting for answer...\n"));
				bRet = fuzzauth_GetAnswerA(sSocket);
			}
		}
		else fuzzauth_Log(LOG_ERR, "fuzzauth_Request() - Sending request to daemon failed: %d: %s\n", ERRORNUMSTR);

		// Fixup endianness in the Request packet.
		g_xPkt_Request.m_nVersion	= ntohl(g_xPkt_Request.m_nVersion);
		g_xPkt_Request.m_nCommand	= ntohl(g_xPkt_Request.m_nCommand);
		g_xPkt_Request.m_nDatatype	= ntohl(g_xPkt_Request.m_nDatatype);
		g_xPkt_Request.m_nLength	= ntohl(g_xPkt_Request.m_nLength);
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_Request() - Unknown command (%d) detected \n", g_xPkt_Request.m_nCommand);
	D(DLOG("fuzzauth_Request(%d) - %s\n", g_xPkt_Request.m_nCommand, bRet ? "Success" : "Failure"));
	return bRet;
}

bool fuzzauth_GetAnswer(SOCKET sSocket, int nDatatype)
{
	bool bRet = false;
	D(DLOG("fuzzauth_GetAnswer(%d)\n", nDatatype));
	bRet = fuzzauth_GetAnswerA(sSocket);
	if (bRet) bRet = fuzzauth_GetAnswerB(sSocket, nDatatype);
	D(DLOG("fuzzauth_GetAnswer(%d) = %s\n", nDatatype, bRet ? "Success" : "Failure"));
	return bRet;
}

void fuzzauth_Disconnect(void)
{
	fuzzauth_DisconnectA((SOCKET *)&g_xServSock, (int *)&g_nConnectCount);
}

void fuzzauth_DisconnectA(SOCKET *sSocket, int *nConnectCount)
{
	D(DLOG("fuzzauth_Disconnect() - Connectcount = %d, Socket=%x\n", g_nConnectCount, *sSocket));
	if (*nConnectCount > 0)
	{
		*nConnectCount = *nConnectCount - 1;
		if (*nConnectCount == 0)
		{
			if (*sSocket != INVALID_SOCKET)
			{
				// Take account of the fact that Request() does a Connect(), Disconnect() sequence.
				*nConnectCount = 42;
				/* Send QUIT message. */
				fuzzauth_InitPackets(c_nCommand_Quit);
				fuzzauth_Request(*sSocket, true);

				/* Close down the socket. */
				closesocket(*sSocket);
				*sSocket = INVALID_SOCKET;
				signal(SIGPIPE, g_xOldSignal);
				*nConnectCount = 0;
			}
			D(else DLOG("fuzzauth_Disconnect() - Connectcount = %d, Socket=%x - but I wasn't connected at all\n", *nConnectCount, *sSocket));
		}
		D(else DLOG("fuzzauth_Disconnect() - Connectcount = %d, Socket=%x - not disconnecting\n", *nConnectCount, sSocket));
	}
	D(else DLOG("fuzzauth_Disconnect() - Connectcount = %d, Socket=%x - successful connect not called before\n", *nConnectCount, *sSocket));
}

bool fuzzauth_Authenticate(const char *sUsername, const char *sPassword)
{
	int bRet = false;
	D(DLOG("fuzzauth_Authenticate()\n"));
	if (fuzzauth_Connect())
	{
		if ((sUsername) && (sPassword))
		{
			fuzzauth_InitPackets(c_nCommand_Authenticate);
			strncpy(g_xPkt_User.m_sUsername, sUsername, sizeof(g_xPkt_User.m_sUsername)-1);
			strncpy(g_xPkt_User.m_sPassword, sPassword, sizeof(g_xPkt_User.m_sPassword)-1);
			if (fuzzauth_Request(g_xServSock, true)) bRet = g_xPkt_Answer.m_nAnswer == c_nAnswer_Granted ? true : false;
		}
		else fuzzauth_Log(LOG_ERR, "fuzzauth_Authenticate() - %s\n", sUsername ? "No Password" : (sPassword ? "No username and no password" : "No Username"));
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_Authenticate() - Couldn't connect...\n");
//	D(DLOG("fuzzauth_Authenticate() - %s\n", bRet ? "Success" : "Failure"));
	return bRet;
}

bool fuzzauth_ChangePassword(char *sUsername, char *sPasswordOld, char *sPasswordNew)
{
	int bRet = false;

	D(DLOG("fuzzauth_ChangePassword()\n"));
	if (fuzzauth_Connect())
	{
		fuzzauth_InitPackets(c_nCommand_ChangePassword);
		strncpy(g_xPkt_User.m_sUsername,		sUsername,		sizeof(g_xPkt_User.m_sUsername)-1);
		strncpy(g_xPkt_User.m_sPassword,		sPasswordOld,	sizeof(g_xPkt_User.m_sPassword)-1);
		strncpy(g_xPkt_User.m_sPasswordNew,	sPasswordNew,	sizeof(g_xPkt_User.m_sPasswordNew)-1);
		if (fuzzauth_Request(g_xServSock, true)) bRet = g_xPkt_Answer.m_nAnswer == c_nAnswer_Granted ? true : false;
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_ChangePassword() - Couldn't connect...\n");
	D(DLOG("fuzzauth_ChangePassword() - %s\n", bRet ? "Success" : "Failure"));
	return bRet;
}

bool fuzzauth_UserExists(const char *sUsername)
{
	struct passwd	xPasswd;
	char				sBuf[1024];
	return fuzzauth_MatchUserName(sUsername, &xPasswd, sBuf, sizeof(sBuf));
}

/* Allocate some space from the nss static buffer.  The buffer and buflen are the pointers passed in by the C library. */
char *fuzzauth_GetStatic(char **sBuffer, size_t *nBuflen, size_t nLength)
{
	char *sResult = NULL;

	/* We return NULL if things aren't set up right, or there isn't enough buffer space left. */
	if ((sBuffer != NULL) && (nBuflen != NULL) && (*nBuflen >= nLength))
	{
	/* Some architectures, like Sparc, need pointers aligned on boundaries */
#if _ALIGNMENT_REQUIRED
		{
			int nMod = nLength % _MAX_ALIGNMENT;
			if (nMod != 0) nLength += _MAX_ALIGNMENT - nMod;
		}
#endif
		sResult = *sBuffer;
		*sBuffer += nLength;
		*nBuflen -= nLength;
	}
	D(else DLOG("fuzzauth_GetStatic() - Setup error: sBuffer = %x, nBuflen = %x, *nBuflen = %d, nLength = %d", sBuffer, nBuflen, nBuflen == NULL ? -1 : *nBuflen, nLength))
	return sResult;
}

bool FA_FillPasswd(struct passwd *xPasswd, char *sBuffer, size_t nBuflen)
{
	bool bRet = false;

	// Clear the complete thing.
	memset(xPasswd, 0, sizeof(struct passwd));
	// g_xPkt comes from fuzzd, so we can trust the data.
	D(DLOG("FA_FillPasswd() - Got the answer... sBuffer = %x, nBufLen = %d\n", sBuffer, nBuflen));
	if ((xPasswd->pw_name		= fuzzauth_GetStatic(&sBuffer, &nBuflen, strlen(g_xPkt_User.m_sUsername)	+ 1))) {
//		D(DLOG("FA_FillPasswd() - Buffer for pw_name = %x...\n", xPasswd->pw_name));
	if ((xPasswd->pw_passwd	= fuzzauth_GetStatic(&sBuffer, &nBuflen, strlen(g_xPkt_User.m_sPassword)		+ 1))) {
//		D(DLOG("FA_FillPasswd() - Buffer for pw_passwd = %x...\n", xPasswd->pw_passwd));
	if ((xPasswd->pw_gecos	= fuzzauth_GetStatic(&sBuffer, &nBuflen, strlen(g_xPkt_User.m_sGECOS)			+ 1))) {
//		D(DLOG("FA_FillPasswd() - Buffer for pw_gecos = %x...\n", xPasswd->pw_gecos));
	if ((xPasswd->pw_dir		= fuzzauth_GetStatic(&sBuffer, &nBuflen, strlen(g_xPkt_User.m_sHomedir)			+ 1))) {
//		D(DLOG("FA_FillPasswd() - Buffer for pw_dir = %x...\n", xPasswd->pw_dir));
	if ((xPasswd->pw_shell	= fuzzauth_GetStatic(&sBuffer, &nBuflen, strlen(g_xPkt_User.m_sShell)			+ 1))) {
//		D(DLOG("FA_FillPasswd() - Buffer for pw_shell = %x...\n", xPasswd->pw_shell));

		D(DLOG("FA_FillPasswd() - Copying the data...\n"));
		strcpy(xPasswd->pw_name,		g_xPkt_User.m_sUsername);
		strcpy(xPasswd->pw_passwd,		g_xPkt_User.m_sPassword);
		xPasswd->pw_uid =					g_xPkt_User.m_nUID;
		xPasswd->pw_gid =					g_xPkt_User.m_nGID;
		strcpy(xPasswd->pw_gecos,		g_xPkt_User.m_sGECOS);
		strcpy(xPasswd->pw_dir,			g_xPkt_User.m_sHomedir);
		strcpy(xPasswd->pw_shell,		g_xPkt_User.m_sShell);
		bRet = true;

	} D(else DLOG("FA_FillPasswd() - Couldn't get buffer for pw_shell\n"));
	} D(else DLOG("FA_FillPasswd() - Couldn't get buffer for pw_dir\n"));
	} D(else DLOG("FA_FillPasswd() - Couldn't get buffer for pw_gecos\n"));
	} D(else DLOG("FA_FillPasswd() - Couldn't get buffer for pw_passwd\n"));
	} D(else DLOG("FA_FillPasswd() - Couldn't get buffer for pw_name - %d\n", strlen(g_xPkt_User.m_sUsername)));
	return bRet;
}

// Get the next user into xPasswd (requires Connect()/Disconnect as start/end pair)
bool fuzzauth_GetUsers(bool bFirst, struct passwd *xPasswd, char *sBuffer, size_t nBuflen)
{
	bool bRet = false;

	D(DLOG("fuzzauth_GetUsers()\n"));
	if (bFirst)
	{
		fuzzauth_InitPackets(c_nCommand_GetUsers);
		if (fuzzauth_Request(g_xServSock, false)) bRet = true;
		else bRet = false;
	}
	else bRet = true;
	if (bRet)
	{
		if (fuzzauth_GetAnswer(g_xServSock, c_nDatatype_User)) bRet = FA_FillPasswd(xPasswd, sBuffer, nBuflen);
		else
		{
			D(DLOG("fuzzauth_GetUsers() - No answer :(\n"));
			bRet = false;
		}
	}
	return bRet;
}

// Match a user on UID
bool fuzzauth_MatchUserID(int nUID, struct passwd *xPasswd, char *sBuffer, size_t nBuflen)
{
	bool bRet = false;

	D(DLOG("fuzzauth_MatchUserID(%d)\n", nUID));
	if (fuzzauth_Connect())
	{
		fuzzauth_InitPackets(c_nCommand_MatchUserID);
		g_xPkt_Number.m_nNumber = nUID;
		if (fuzzauth_Request(g_xServSock, false))
		{
			if (fuzzauth_GetAnswer(g_xServSock, c_nDatatype_User)) bRet = FA_FillPasswd(xPasswd, sBuffer, nBuflen);
		}
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_MatchUserID() - Couldn't connect...\n");
	return bRet;
}

// Match a user on Name
bool fuzzauth_MatchUserName(const char *sName, struct passwd *xPasswd, char *sBuffer, size_t nBuflen)
{
	bool bRet = false;

	D(DLOG("fuzzauth_MatchUserName('%s')\n", sName));
	if (fuzzauth_Connect())
	{
		fuzzauth_InitPackets(c_nCommand_MatchUserName);
		strncpy(g_xPkt_String.m_sString, sName, sizeof(g_xPkt_String.m_sString)-1);
		if (fuzzauth_Request(g_xServSock, false))
		{
			if (fuzzauth_GetAnswer(g_xServSock, c_nDatatype_User)) bRet = FA_FillPasswd(xPasswd, sBuffer, nBuflen);
		}
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_MatchUserName() - Couldn't connect...\n");
	return bRet;
}

int FA_FillGroup(struct group *xGroup, char *sBuffer, size_t nBuflen)
{
	int nRet = c_nAnswer_NoMoreData;

	// Clear the complete thing.
	memset(xGroup, 0, sizeof(struct group));
	// g_xPkt comes from fuzzd, so we can trust the data.
//	D(DLOG("FA_FillGroup() - Got the answer... sBuffer = %x, nBufLen = %d\n", sBuffer, nBuflen));
	if ((xGroup->gr_name		= fuzzauth_GetStatic(&sBuffer, &nBuflen, strlen(g_xPkt_Group.m_sGroupname)	+ 1)))
	{
//		D(DLOG("FA_FillGroup() - Buffer for gr_name = %x...\n", xGroup->gr_name));
		if ((xGroup->gr_passwd	= fuzzauth_GetStatic(&sBuffer, &nBuflen, strlen(g_xPkt_Group.m_sPassword)	+ 1)))
		{
			unsigned int nMemberCount = g_xPkt_Group.m_nNumMembers, i = 0;
//			D(DLOG("FA_FillGroup() - Buffer for gr_passwd = %x...\n", xGroup->gr_passwd));
			D(DLOG("FA_FillGroup() - Copying the data...\n"));
			strcpy(xGroup->gr_name,		g_xPkt_Group.m_sGroupname);
			strcpy(xGroup->gr_passwd,	g_xPkt_Group.m_sPassword);
			xGroup->gr_gid =				g_xPkt_Group.m_nGID;

			// Allocate space for the users ((x+1)*4 bytes on 32bit systems...).
			if ((xGroup->gr_mem = (char **)fuzzauth_GetStatic(&sBuffer, &nBuflen, (nMemberCount+1) * sizeof(char *))))
			{
				// And get the users belonging to this group.
				while ((i < nMemberCount) && (fuzzauth_GetAnswerA(g_xServSock)) && (g_xPkt_Answer.m_nAnswer == c_nAnswer_Granted) && (g_xPkt_Answer.m_nDatatype == c_nDatatype_String))
				{
					if (fuzzauth_GetAnswerB(g_xServSock, c_nDatatype_String))
					{
						if ((xGroup->gr_mem[i] = fuzzauth_GetStatic(&sBuffer, &nBuflen, strlen(g_xPkt_String.m_sString) + 1)))
						{
							strcpy(xGroup->gr_mem[i], g_xPkt_String.m_sString);
							D(DLOG("FA_FillGroup() - Added user %d: %s", i, xGroup->gr_mem[i]));
							// Next please...
							i++;
						}
						D(else DLOG("FA_FillGroup() - No buffer left for user."));
					}
					D(else DLOG("FA_FillGroup() - Wrong answer received."));
				}
				D(DLOG("FA_FillGroup() - Terminating members at %d.", i));
				xGroup->gr_mem[i] = NULL;
				D(for (i = 0; i < nMemberCount + 1; i++) { DLOG("FA_FillGroup() - Dump: %d = \"%s\".\n", i, xGroup->gr_mem[i]); } )
				nRet = c_nAnswer_Granted;
			}
			else
			{
				D(DLOG("FA_FillGroup() - Couldn't allocate memory for members."));
				nRet = c_nAnswer_NeedDataPacket;
			}
		}
		D(else DLOG("FA_FillGroup() - Couldn't get buffer for gr_passwd\n"));
	}
	D(else DLOG("FA_FillGroup() - Couldn't get buffer for gr_name - %d\n", strlen(g_xPkt_Group.m_sGroupname)));
	return nRet;
}

// Get the next group into xGroup (requires Connect()/Disconnect as start/end pair)
int fuzzauth_GetGroups(bool bFirst, struct group *xGroup, char *sBuffer, size_t nBuflen)
{
	int nRet = c_nAnswer_NoMoreData;
	bool bRet = false;

	D(DLOG("fuzzauth_GetGroups()\n"));
	if (bFirst)
	{
		fuzzauth_InitPackets(c_nCommand_GetGroups);
		if (fuzzauth_Request(g_xServSock, false)) bRet = true;
		else bRet = false;
	}
	else bRet = true;
	if (bRet)
	{
		bRet = fuzzauth_GetAnswerA(g_xServSock);
		// Check if we got a correct answer.
		if ((bRet) && (g_xPkt_Answer.m_nAnswer == c_nAnswer_Granted) && (g_xPkt_Answer.m_nDatatype == c_nDatatype_Group))
		{
			// Get the group info.
			if (fuzzauth_GetAnswerB(g_xServSock, c_nDatatype_Group))
			{
				nRet = FA_FillGroup(xGroup, sBuffer, nBuflen);
			}
			D(else DLOG("fuzzauth_GetGroups() - Wrong answer received."));
		}
		else
		{
			bRet = false;
			D(if (g_xPkt_Answer.m_nAnswer != c_nAnswer_NoMoreData) DLOG("fuzzauth_GetGroups() - Wrong answer or datatype received."));
		}
	}
	return nRet;
}

// Match a group on ID
bool fuzzauth_MatchGroupID(int nGID, struct group *xGroup, char *sBuffer, size_t nBuflen)
{
	bool bRet = false;

	D(DLOG("fuzzauth_MatchGroupID('%d')\n", nGID));
	if (fuzzauth_Connect())
	{
		fuzzauth_InitPackets(c_nCommand_MatchGroupID);
		g_xPkt_Number.m_nNumber = nGID;
		if (fuzzauth_Request(g_xServSock, false))
		{
			if (fuzzauth_GetAnswer(g_xServSock, c_nDatatype_Group))
			{
				bRet = FA_FillGroup(xGroup, sBuffer, nBuflen);
			}
			else bRet = false;
		}
		else bRet = false;
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_MatchGroupID() - Couldn't connect...\n");
	return bRet;
}

// Match a group on Name
bool fuzzauth_MatchGroupName(const char *sName, struct group *xGroup, char *sBuffer, size_t nBuflen)
{
	bool bRet = false;

	D(DLOG("fuzzauth_MatchGroupName('%s')\n", sName));
	if (fuzzauth_Connect())
	{
		fuzzauth_InitPackets(c_nCommand_MatchGroupName);
		strncpy(g_xPkt_String.m_sString, sName, sizeof(g_xPkt_String.m_sString)-1);
		if (fuzzauth_Request(g_xServSock, false))
		{
			if (fuzzauth_GetAnswer(g_xServSock, c_nDatatype_Group))
			{
				bRet = FA_FillGroup(xGroup, sBuffer, nBuflen);
			}
			else bRet = false;
		}
		else bRet = false;
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_MatchGroupName() - Couldn't connect...\n");
	return bRet;
}

unsigned int UserToUID(char *sUsername)
{
	int nRet = atoi(sUsername);
	if (nRet == 0)
	{
		struct passwd	xPassword;
		char				sBuf[1024];
		if (fuzzauth_MatchUserName(sUsername, &xPassword, sBuf, sizeof(sBuf)))
		{
			nRet = xPassword.pw_uid;
		}
		else nRet = 0;
	}
	return nRet;
}

unsigned int GroupToGID(char *sGroupname)
{
	int nRet = atoi(sGroupname);
	if (nRet == 0)
	{
		struct group	xGroup;
		char				sBuf[1024];
		if (fuzzauth_MatchGroupName(sGroupname, &xGroup, sBuf, sizeof(sBuf)))
		{
			nRet = xGroup.gr_gid;
		}
		else nRet = 0;
	}
	return nRet;
}

bool fuzzauth_UserAdd(char *sUsername, char *sPassword, char *sPrimaryGroup, char *sGECOS, char *sHomedir, char *sShell)
{
	bool bRet = false;

	D(DLOG("fuzzauth_UserAdd(\"%s\")\n", sUsername));
	if (fuzzauth_Connect())
	{
		int nUID = -1, nGID = -1;
		if (sPrimaryGroup) nGID = GroupToGID(sPrimaryGroup);
		fuzzauth_InitPackets(c_nCommand_UserAdd);
		g_xPkt_User.m_nUID = nUID;
		g_xPkt_User.m_nGID = nGID;
		strncpy(g_xPkt_User.m_sUsername,			sUsername,	sizeof(g_xPkt_User.m_sUsername)-1);
		strncpy(g_xPkt_User.m_sPassword,			sPassword,	sizeof(g_xPkt_User.m_sPassword)-1);
		strncpy(g_xPkt_User.m_sPasswordNew,		sPassword,	sizeof(g_xPkt_User.m_sPasswordNew)-1);
		strncpy(g_xPkt_User.m_sGECOS,		 		sGECOS,		sizeof(g_xPkt_User.m_sGECOS)-1);
		strncpy(g_xPkt_User.m_sHomedir,	 		sHomedir,	sizeof(g_xPkt_User.m_sHomedir)-1);
		strncpy(g_xPkt_User.m_sShell,		 		sShell,		sizeof(g_xPkt_User.m_sShell)-1);

		if (fuzzauth_Request(g_xServSock, true))
		{
			D(DLOG("fuzzauth_UserAdd(\"%s\") - Success\n", sUsername));
			bRet = true;
		}
		else bRet = false;
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_UserAdd() - Couldn't connect...\n");
	return bRet;
}

bool fuzzauth_UserRemove(char *sUsername)
{
	bool bRet = false;

	D(DLOG("fuzzauth_UserRemove(\"%s\")\n", sUsername));
	if (fuzzauth_Connect())
	{
		fuzzauth_InitPackets(c_nCommand_UserRemove);
		strncpy(g_xPkt_String.m_sString,	sUsername,	sizeof(g_xPkt_String.m_sString)-1);
		if (fuzzauth_Request(g_xServSock, true))
		{
			D(DLOG("fuzzauth_UserRemove(\"%s\") - Success\n", sUsername));
			bRet = true;
		}
		else bRet = false;
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_UserRemove() - Couldn't connect...\n");
	return bRet;
}

bool fuzzauth_UserChange(char *sUsername, char *sNewUsername, char *sPassword)
{
	bool bRet = false;
	return bRet;
}

bool fuzzauth_GroupAdd(char *sGroupname, char *sPassword)
{
	bool bRet = false;
	return bRet;
}

bool fuzzauth_GroupRemove(char *sGroupname)
{
	bool bRet = false;

	D(DLOG("fuzzauth_GroupRemove(\"%s\")\n", sGroupname));
	if (fuzzauth_Connect())
	{
		fuzzauth_InitPackets(c_nCommand_GroupRemove);
		strncpy(g_xPkt_String.m_sString,	sGroupname,	sizeof(g_xPkt_String.m_sString)-1);
		if (fuzzauth_Request(g_xServSock, true))
		{
			D(DLOG("fuzzauth_GroupRemove(\"%s\") - Success\n", sGroupname));
			bRet = true;
		}
		else bRet = false;
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_GroupRemove() - Couldn't connect...\n");
	return bRet;
}

bool fuzzauth_GroupChange(char *sGroupname, char *sNewGroupname, char *sPassword)
{
	bool bRet = false;
	return bRet;
}

bool fuzzauth_MemberAdd(char *sUsername, char *sGroupname)
{
	bool bRet = false;

	D(DLOG("fuzzauth_MemberAdd(\"%s:%s\")\n", sUsername, sGroupname));
	if (fuzzauth_Connect())
	{
		fuzzauth_InitPackets(c_nCommand_MemberAdd);
		g_xPkt_Number2.m_nNumber1 = UserToUID(sUsername);
		g_xPkt_Number2.m_nNumber2 = GroupToGID(sGroupname);
		if ((g_xPkt_Number2.m_nNumber1 != 0) && (g_xPkt_Number2.m_nNumber2 != 0))
		{
			if (fuzzauth_Request(g_xServSock, true))
			{
				D(DLOG("fuzzauth_MemberAdd(\"%s:%s\") - Success\n", sUsername, sGroupname));
				bRet = true;
			}
		}
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_MemberAdd() - Couldn't connect...\n");
	return bRet;
}

bool fuzzauth_MemberRemove(char *sUsername, char *sGroupname)
{
	bool bRet = false;

	D(DLOG("fuzzauth_MemberRemove(\"%s:%s\")\n", sUsername, sGroupname));
	if (fuzzauth_Connect())
	{
		fuzzauth_InitPackets(c_nCommand_MemberRemove);
		g_xPkt_Number2.m_nNumber1 = UserToUID(sUsername);
		g_xPkt_Number2.m_nNumber2 = GroupToGID(sGroupname);
		if ((g_xPkt_Number2.m_nNumber1 != 0) && (g_xPkt_Number2.m_nNumber2 != 0))
		{
			if (fuzzauth_Request(g_xServSock, true))
			{
				D(DLOG("fuzzauth_MemberRemove(\"%s:%s\") - Success\n", sUsername, sGroupname));
				bRet = true;
			}
		}
		fuzzauth_Disconnect();
	}
	else fuzzauth_Log(LOG_ERR, "fuzzauth_MemberRemove() - Couldn't connect...\n");
	return bRet;
}
