/*	FUZZauth - http://unfix.org/projects/fuzzauth & http://fuzzauth.sourceforge.net
 * FUZZauth_config.c by Jeroen Massar <jeroen@unfix.org>
 *
 * Config functions
*/

#include "fuzzauth_config.h"

#define __STRICT_ANSI__ 1
#define _GNU_SOURCE 1
#define _LARGEFILE64_SOURCE 1
#define _REENTRANT 1

#include <features.h>

#ifndef __GNUC__
#define __attribute__(x) /*nothing*/
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef DEBUG
#define D(x) x
#include <syslog.h>
void DLOG(const char *sFormat, ... )
{
	va_list xArgs;
	va_start(xArgs, sFormat);
	openlog(FUZZAUTH_NAME"-Config", LOG_PID, LOG_AUTH);
	vsyslog(LOG_DEBUG, sFormat, xArgs);
	closelog();
	va_end(xArgs);
}
#else
#define D(x)
#endif

char *ParseConfig(char *sConfigname, struct Config *xConfig, int nConfigEntries)
{
	FILE			*xFile = NULL;
	char			*sFilename = NULL;
	char			*sError = NULL;
	struct stat	xStat;

	D(DLOG("ParseConfig() - %s, entries = %d\n", sConfigname, nConfigEntries));

	asprintf(&sFilename, "%s/etc/%s.conf", FUZZAUTH_HOMEDIR, sConfigname);
	if (sFilename)
	{
		if (lstat(sFilename, &xStat) == 0)
		{
			if (S_ISREG(xStat.st_mode))
			{
				if (((xStat.st_mode & S_ISUID) == 0) && ((xStat.st_mode & S_ISGID) == 0))
				{
					if ((xStat.st_mode & S_ISVTX) == 0)
					{
						if ((xStat.st_mode & 0777) == 0750)
						{
							if ((xFile = fopen(sFilename, "r")))
							{
								char	sBuf[4096];
								int	nLen, i, j, nFoundOptionName, nBegin, nEnd;
								bool bQuotesDouble, bQuotesSingle;

								// Unless an unknown option is encountered we are ok with this.
								while ((sError == NULL) && (!feof(xFile)))
								{
									sBuf[0] = '\0';
									if (fgets(sBuf, sizeof(sBuf), xFile) == sBuf)
									{
										nFoundOptionName = -1;
										nBegin = -1;
										nEnd = -1;
										bQuotesDouble = false;
										bQuotesSingle = false;

										nLen = strlen(sBuf);
										if ((nLen > 0) && (sBuf[nLen-1] == '\n'))
										{
											nLen--;
											sBuf[nLen] = '\0';
										}
										D(DLOG("Line \"%s\"\n", sBuf));
										if ((sBuf[0] == '\0') || (sBuf[0] == '#') || ((sBuf[0] == '/') && (sBuf[1] == '/')))
										{
											D(DLOG("Ignoring comment line...\n"));
										}
										else
										{
											i = 0;
											while ((sError == NULL) && (nEnd == -1) && (i < nLen))
											{
												// If we haven't got an optionname, just skip along till...
												if (nFoundOptionName == -1)
												{
													// Mark the end of the optionname when we meet:
													// - space(' ')
													// - tab ('\t')
													// - equalsign ('=') in between optionname and value.
													if ((sBuf[i] == ' ') || (sBuf[i] == '\t') || (sBuf[i] == '='))
													{
														// Compare Optionnames with the ones in the config.
														for (j=0; (nFoundOptionName == -1) && (j < nConfigEntries); j++)
														{
															D(DLOG("Comparing \"%.*s\" with \"%s\" (i=%d/e=%d)\n", i, sBuf, xConfig[j].m_sName, j, nConfigEntries));
															if (strncasecmp(xConfig[j].m_sName, sBuf, i) == 0) nFoundOptionName = j;
														}
														if (nFoundOptionName == -1)
														{
															asprintf(&sError, "Unknown option \"%s\" encountered.\n", sBuf);
														}
														D(else DLOG("Encountered Option \"%s\"\n", xConfig[nFoundOptionName].m_sName));
													}
												}
												else
												{
													// Skip whitespace at the start.
													if (nBegin == -1)
													{
														D(DLOG("Checking '%c' for begin...\n", (int)sBuf[i]));
														// Mark the begin of the value when we do not meet:
														// - space(' ')
														// - tab ('\t')
														// - equalsign ('=') in between optionname and value.
														if ((sBuf[i] != ' ') && (sBuf[i] != '\t') && (sBuf[i] != '='))
														{
															if (sBuf[i] == '\'')
															{
																D(DLOG("It's a single quote!\n"));
																bQuotesSingle = true;
															}
															else if (sBuf[i] == '"')
															{
																D(DLOG("It's a double quote!\n"));
																bQuotesDouble = true;
															}
															else
															{
																D(DLOG("BEGIN found...\n"));
																nBegin = i;
																nEnd = -1;
															}
														}
														D(else DLOG("It's whitespace...\n"));
													}
													else
													{
														D(DLOG("Checking '%c' for end...\n", (int)sBuf[i]));
														if (bQuotesSingle)
														{
															if (sBuf[i] == '\'') nEnd = i-1;
														}
														else if (bQuotesDouble)
														{
															if (sBuf[i] == '"') nEnd = i-1;
														}
														else if ((sBuf[i] == ' ') || (sBuf[i] == '\t'))
														{
															nEnd = i;
														}
														D(else DLOG("None of the above...\n"));
													}
												}
												i++;
											}
											// If there where no quotes we probably run till the end of the line
											if ((nBegin != -1) && (!bQuotesSingle) && (!bQuotesDouble))
											{
												nEnd = i;
											}
											if (nFoundOptionName == -1)
											{
												asprintf(&sError, "Option \"%s\" is unknown in configfile \"%s\".\n", sBuf, sFilename);
											}
											else if (nEnd == -1)
											{
												asprintf(&sError, "Option \"%s\" was given no value in configfile \"%s\".\n", xConfig[nFoundOptionName].m_sName, sFilename);
											}
											else
											{
												if ((xConfig[nFoundOptionName].m_sValue = malloc(i - nBegin + 1)))
												{
													strncpy(xConfig[nFoundOptionName].m_sValue, &sBuf[nBegin], nEnd-nBegin+1);
													D(DLOG("Option \"%s\" set to \"%s\"\n", xConfig[nFoundOptionName].m_sName, xConfig[nFoundOptionName].m_sValue));
												}
											}
										}
									}
								}
								fclose(xFile);
							}
						}
						else asprintf(&sError, "File mode for %s should be %o and is %o.\n", sFilename, 0750, xStat.st_mode & 0x1FF);
					}
					else asprintf(&sError, "Configfile \"%s\" has sticky bit set.\n", sFilename);
				}
				else asprintf(&sError, "Configfile \"%s\" has suid bits set.\n", sFilename);
			}
			else asprintf(&sError, "Configfile \"%s\" is not a regular file.\n", sFilename);
			free(sFilename);
		}
		else asprintf(&sError, "Unable to stat configfile \"%s\".\n", sFilename);
	}
	else asprintf(&sError, "Config filename couldn't be generated.\n");
	D(if (sError) DLOG("Encountered error: %s\n", sError));
	return sError;
}

void FreeConfig(struct Config *xConfig, int nConfigEntries)
{
	int i;
	for (i=0; i<nConfigEntries; i++)
	{
		if ((xConfig[i].m_sValue) && (xConfig[i].m_nValueSize > 0))
		{
			free(xConfig[i].m_sValue);
			xConfig[i].m_sValue = NULL;
			xConfig[i].m_nValueSize = 0;
		}
	}
}
