/*	FUZZauth - http://unfix.org/projects/fuzzauth & http://fuzzauth.sourceforge.net
 * FUZZ.h - common definitions for the FUZZd (FUZZauth Daemon) and fuzzauth_* authenticator libraries by Jeroen Massar <jeroen@unfix.org>
*/

#ifndef FUZZD_H
#define FUZZD_H "FUZZauth"

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#include <ws2tcpip.h>
#include <tpipv6.h>  // For IPv6 Tech Preview.
#define ERRORNUMSTR WSAGetLastError(),DecodeError(WSAGetLastError())
#define CANNOT_DAEMONIZE
#else
#define SOCKET int
#define INVALID_SOCKET (-1)
#define SOCKET_ERROR (-1)
#define closesocket close
#define ERRORNUMSTR errno,strerror(errno)
#endif

#define __STRICT_ANSI__ 1
#define _GNU_SOURCE 1
#define _LARGEFILE64_SOURCE 1
#define _REENTRANT 1

#include <features.h>

#ifndef __GNUC__
#define __attribute__(x) /*nothing*/
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/types.h>
#include <fcntl.h>
#include <crypt.h>
#include <dlfcn.h>
#include <time.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>

#ifndef WIN32
#include <unistd.h>
#include <netdb.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <signal.h>
#endif
#include <sys/un.h>
#include <syslog.h>

/* FUZZauth */
#include "fuzzauth.h"

#define FUZZD_NAME				"FUZZd"							// The name of this program.

/* Configurable items */
#define DEFAULT_FAMILY			PF_UNSPEC						// Accept either IPv4 or IPv6.
#define BUFFER_SIZE				1024								// Should be more than sufficient for most messages.

/* Internal Structures */
#include "list.h"

struct Client
{
	SOCKET							m_xSocket;						// The socket to which the client is connected.
	struct sockaddr_storage		m_xFrom;							// Where the client came from.
	bool								m_bUnix;							// This is a UNIX Socket.
	char								m_sHostname[NI_MAXHOST];	// The name of the client.
	char								m_sBuffer[BUFFER_SIZE];		// Buffer in which we put conversation data.
	unsigned int					m_nOffset;						// Offset in buffer (for handling fragmented messages).
	enum c_nCommand				m_nMode;							// Command Mode.
	bool								m_bGotCommand;					// Got a command.
	struct faPacket_Request		m_xPkt_Request;				// Request Packet.
	struct faPacket_Answer		m_xPkt_Answer;					// Answer Packet.
	struct faPacket_Number		m_xPkt_Number;					// Number Packet.
	struct faPacket_Number2		m_xPkt_Number2;				// Number2 Packet.
	struct faPacket_String		m_xPkt_String;					// String Packet.
	struct faPacket_User			m_xPkt_User;					// User Packet.
	struct faPacket_Group		m_xPkt_Group;					// Group Packet.
};

struct Authenticator
{
	bool	(*cmd_Initialize)(void);								// Initialize theplugin (connect to db etc..)
	bool	(*cmd_Shutdown)(void);									// Shutdown the plugin (disconnect etc...)
	bool	(*cmd_GetInfo)(char *sBuf, int nSize);				// Get some information about this authenticator.
	bool	(*cmd_Authenticate	)(struct Client *xClient);	// Authenticate a User.
	bool	(*cmd_ChangePassword	)(struct Client *xClient);	// Change the password of a User.
	bool	(*cmd_GetUsers			)(struct Client *xClient);	// Get a list of all users.
	bool	(*cmd_MatchUserID		)(struct Client *xClient);	// Match User by ID.
	bool	(*cmd_MatchUserName	)(struct Client *xClient); // Match User by Name.
	bool	(*cmd_MatchUserGECOS	)(struct Client *xClient); // Match User by GECOS (substring match).
	bool	(*cmd_GetGroups		)(struct Client *xClient); // Get a list of all groups (inc. members).
	bool	(*cmd_MatchGroupID	)(struct Client *xClient); // Match Group by ID.
	bool	(*cmd_MatchGroupName	)(struct Client *xClient); // Match Group by Name.
	bool	(*cmd_UserAdd			)(struct Client *xClient); // Add a user.
	bool	(*cmd_UserRemove		)(struct Client *xClient); // Remove a user.
	bool	(*cmd_UserChange		)(struct Client *xClient); // Change a user.
	bool	(*cmd_GroupAdd			)(struct Client *xClient); // Add a group.
	bool	(*cmd_GroupRemove		)(struct Client *xClient); // Remove a group.
	bool	(*cmd_GroupChange		)(struct Client *xClient); // Change a group.
	bool	(*cmd_MemberAdd		)(struct Client *xClient); // Add a member.
	bool	(*cmd_MemberRemove	)(struct Client *xClient); // Remove a member.
};

/* Debug Interface */
#ifdef DEBUG
#define D(x) x
void DLOG(const char *sFormat, ... ) __attribute__((format(printf,1,2)));
#else
#define D(x) /* nothing */
#endif

/* Functions to be used by fuzzd_auth_* */
void fuzzd_LogA(int nPriority, const char *sFormat, va_list xArgs) __attribute__((format(printf,2,0)));
void fuzzd_Log(int nPriority, const char *sFormat, ... ) __attribute__((format(printf,2,3)));

bool fuzzd_ReturnAnswer(struct Client *xClient);
char *fuzzd_Encrypt(char *sPassword);
bool fuzzd_ComparePassword(char *sStored, char *sInput);

#ifdef FA_CRYPT_MHASH
#include <mhash.h>
#endif

#endif // ifndef FUZZD_H
