/* libnss_fuzzauth.so by Jeroen Massar <jeroen@unfix.org> */

/* 'info libc "NSS Module Function Internals"' */

/*
	We provide:

	passwd
	group
*/

/* NSS */
#include <nss.h>

/* Password & Group */
#include <pwd.h>
#include <grp.h>

/* FUZZauth */
#include "fuzzauth_common.h"

volatile bool g_bUserFirst = false;
volatile bool g_bGroupFirst = false;

/*********************************************************
 NSS Passwd (pw) Functions
*********************************************************/

/* Open our connection to the authentication deamon */
enum nss_status _nss_fuzzauth_setpwent(void)
{
	D(DLOG("setpwent\n"));
	g_bUserFirst = true;
	return fuzzauth_Connect() ? NSS_STATUS_SUCCESS : NSS_STATUS_TRYAGAIN;
}

/* Close our connection to the authentication deamon */
enum nss_status _nss_fuzzauth_endpwent(void)
{
	D(DLOG("endpwent\n"));
	fuzzauth_Disconnect();
	g_bUserFirst = false;
	return NSS_STATUS_SUCCESS;
}

/* Fetch the next password entry from password database */
enum nss_status _nss_fuzzauth_getpwent_r(struct passwd *result, char *buffer, size_t buflen, int *errnop)
{
	enum nss_status	nRet;

	D(DLOG("getpwent_r\n"));
	if (fuzzauth_GetUsers(g_bUserFirst, result, buffer, buflen))
	{
		nRet = NSS_STATUS_SUCCESS;
		g_bUserFirst = false;
	}
	else nRet = NSS_STATUS_NOTFOUND;
	*errnop = 0;
	return nRet;
}

/* Return passwd struct from uid */
enum nss_status _nss_fuzzauth_getpwuid_r(uid_t uid, struct passwd *result, char *buffer, size_t buflen, int *errnop)
{
	enum nss_status nRet = NSS_STATUS_NOTFOUND;

	D(DLOG("getpwuid_r() - Searching for %d\n", uid));
	if (fuzzauth_MatchUserID(uid, result, buffer, buflen))
	{
		D(DLOG("getpwnam_r() - Found ya!\n"));
		nRet = NSS_STATUS_SUCCESS;
	}
	*errnop = 0;
	D(DLOG("getpwnam_r() - Returning %d (SUCCESS=%d, NOTFOUND=%d)\n", nRet, NSS_STATUS_SUCCESS, NSS_STATUS_NOTFOUND));
	return nRet;
}

/* Return passwd struct from username */
enum nss_status _nss_fuzzauth_getpwnam_r(const char *name, struct passwd *result, char *buffer, size_t buflen, int *errnop)
{
	enum nss_status nRet = NSS_STATUS_NOTFOUND;

	D(DLOG("getpwnam_r() - Searching for %s\n", name));
	if (fuzzauth_MatchUserName(name, result, buffer, buflen))
	{
		D(DLOG("getpwnam_r() - Found ya!\n"));
		nRet = NSS_STATUS_SUCCESS;
	}
	if (errnop) *errnop = 0;
	D(DLOG("getpwnam_r() - Returning %d (SUCCESS=%d, NOTFOUND=%d)\n", nRet, NSS_STATUS_SUCCESS, NSS_STATUS_NOTFOUND));
	return nRet;
}

/*********************************************************
 NSS Group (gr) Functions
*********************************************************/

/* Open our connection to the authentication deamon */
enum nss_status _nss_fuzzauth_setgrent(void)
{
	D(DLOG("setgrent\n"));
	g_bGroupFirst = true;
	return fuzzauth_Connect() ? NSS_STATUS_SUCCESS : NSS_STATUS_TRYAGAIN;
}

/* Close our connection to the authentication deamon */
enum nss_status _nss_fuzzauth_endgrent(void)
{
	D(DLOG("endgrent\n"));
	fuzzauth_Disconnect();
	g_bGroupFirst = false;
	return NSS_STATUS_SUCCESS;
}

char *members = { NULL };

/* Fetch the next password entry from password database */
enum nss_status _nss_fuzzauth_getgrent_r(struct group *result, char *buffer, size_t buflen, int *errnop)
{
	int					nAns;
	enum nss_status	nRet;

	*errnop = 0;
	D(DLOG("getgrent_r\n"));
	nAns = fuzzauth_GetGroups(g_bGroupFirst, result, buffer, buflen);
	if (nAns == c_nAnswer_Granted)
	{
		nRet = NSS_STATUS_SUCCESS;
		g_bGroupFirst = false;
	}
	else if (nAns == c_nAnswer_NeedDataPacket) nRet = NSS_STATUS_TRYAGAIN;
	else nRet = NSS_STATUS_NOTFOUND;
	return nRet;
}

/* Return passwd struct from gid */
enum nss_status _nss_fuzzauth_getgrgid_r(uid_t gid, struct group *result, char *buffer, size_t buflen, int *errnop)
{
	enum nss_status nRet = NSS_STATUS_NOTFOUND;

	*errnop = 0;
	D(DLOG("getgruid_r() - Searching for %d\n", gid));
	if (fuzzauth_MatchGroupID(gid, result, buffer, buflen))
	{
		nRet = NSS_STATUS_SUCCESS;
	}
	return nRet;
}

/* Return passwd struct from username */
enum nss_status _nss_fuzzauth_getgrnam_r(const char *name, struct group *result, char *buffer, size_t buflen, int *errnop)
{
	enum nss_status nRet = NSS_STATUS_NOTFOUND;

	*errnop = 0;
	D(DLOG("getgrnam_r() - Searching for %s\n", name));
	if (fuzzauth_MatchGroupName(name, result, buffer, buflen))
	{
		nRet = NSS_STATUS_SUCCESS;
	}
	return nRet;
}
