/* libpam_fuzzauth.so by Jeroen Massar <jeroen@unfix.org> */

/* FUZZauth */
#include "fuzzauth_common.h"

/************************
 PAM Includes
************************/
#if defined(SUNOS5) || defined(SUNOS4)
/* Solaris always uses dynamic pam modules */
#define PAM_EXTERN extern
#include <security/pam_appl.h> 
#define PAM_AUTHTOK_RECOVER_ERR PAM_AUTHTOK_RECOVERY_ERR
#endif
#define PAM_SM_AUTH FUZZAUTH_NAME
#define PAM_SM_SESSION FUZZAUTH_NAME
#define PAM_SM_ACCOUNT FUZZAUTH_NAME
#define PAM_SM_PASSWORD FUZZAUTH_NAME
#include <security/pam_modules.h>

static void PAM_drop_reply(struct pam_response *xReply, int nReplies)
{
	int i;
	for (i=0; i<nReplies; i++)
	{
		if (xReply[i].resp)
		{
			memset(xReply[i].resp, 0, strlen(xReply[i].resp));
			free(xReply[i].resp);
		}
	}
	if (xReply) free(xReply);
}

static int PAM_Converse(pam_handle_t *pamHandle, int nArgs, struct pam_message **xMessage, struct pam_response **xResponse)
{
	int					nRet;
	struct pam_conv 	*xConv;

	D(DLOG("PAM_Converse()\n"));
	nRet = pam_get_item(pamHandle, PAM_CONV, (const void **)&xConv);
	if (nRet == PAM_SUCCESS)
	{
		D(DLOG("PAM_Converse() - Conversating...\n"));
		nRet = xConv->conv(nArgs, (const struct pam_message **)xMessage, xResponse, xConv->appdata_ptr);
	}
	else fuzzauth_Log(LOG_ERR, "PAM_Converse() - Couldn't get PAM conversation item.\n");
	D(DLOG("PAM_Converse() - %s\n", nRet == PAM_SUCCESS ? "Success" : "Failure"));
	return nRet;
}

static int PAM_Conversate(pam_handle_t *pamHandle)
{
	struct pam_message	xMsg, *pMsg;
	struct pam_response	*xResp;
	int						nRet;
	char						*sToken = NULL;

	D(DLOG("PAM_Conversate()\n"));
	pMsg = &xMsg;
	xMsg.msg_style = PAM_PROMPT_ECHO_OFF;
	xMsg.msg = "Password: ";

	/* Call the conversation expecting 1 response. */
	xResp = NULL;
	nRet = PAM_Converse(pamHandle, 1, &pMsg, &xResp);
	if (xResp != NULL)
	{
		const char *sItem;
		/* interpret the response */
		if (nRet == PAM_SUCCESS)
		{
			/* a good conversation */
			sToken = xResp[0].resp?strdup(xResp[0].resp):NULL;
			if (sToken == NULL)
			{
				nRet = PAM_AUTHTOK_RECOVER_ERR;
				D(DLOG("PAM_Conversate() - Authentication Token is NULL\n"));
			}
			else
			{
				D(DLOG("PAM_Conversate() - Setting Authentication Token\n"));
				/* set the auth token */
				nRet = pam_set_item(pamHandle, PAM_AUTHTOK, sToken);
				memset(sToken, 0, strlen(sToken));
				free(sToken);
				if (!((nRet != PAM_SUCCESS) || (nRet = pam_get_item(pamHandle, PAM_AUTHTOK, (const void **)&sItem)) != PAM_SUCCESS))
				{
					PAM_drop_reply(xResp, 1);
				}
			}
		}
		D(else DLOG("PAM_Conversate() - Got a response but with %d as an error\n", nRet));
	}
	else
	{
		nRet = (nRet == PAM_SUCCESS) ? PAM_AUTHTOK_RECOVER_ERR : nRet;
		D(DLOG("PAM_Conversate() - No Response\n"));
	}
	D(DLOG("PAM_Conversate() - %s\n", nRet == PAM_SUCCESS ? "Success" : "Failure"));
	return nRet;
}

/************************
 PAM Interface Functions
************************/
PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamHandle, int nFlags, int nArgc, const char **sArgv)
{
	const char	*sUsername;
	const char	*sPassword;
	int			nRet = PAM_AUTH_ERR;

	/* Get the username */
	D(DLOG("pam_sm_authenticate()\n"));
	nRet = pam_get_user(pamHandle, &sUsername, NULL);
	if (nRet == PAM_SUCCESS)
	{
		if (sUsername)
		{
			// Converse just to be sure we have the password
			nRet = PAM_Conversate(pamHandle);
			if (nRet == PAM_SUCCESS)
			{
				/* Get the password */
				nRet = pam_get_item(pamHandle,PAM_AUTHTOK,(const void **)&sPassword);
				if (nRet == PAM_SUCCESS)
				{
					D(DLOG("pam_sm_authenticate() - Verify user `%s' with password `%s'\n", sUsername, sPassword ? "xxx" : "!none!"));
					/* Now use the username to look up password */
					if (fuzzauth_Authenticate(sUsername, sPassword)) nRet = PAM_SUCCESS;
					else nRet = PAM_AUTH_ERR;
				}
			}
			else fuzzauth_Log(LOG_ERR, "pam_sm_authenticate() - Couldn't get password from PAM\n");
		}
		else nRet = PAM_SERVICE_ERR;
	}
	D(DLOG("pam_sm_authenticate() - %s: %s\n", sUsername, nRet == PAM_SUCCESS ? "Success" : "Failure"));
	return nRet;
	
	// Unused Parameters
	pamHandle = pamHandle;
	nFlags = nFlags;
	nArgc = nArgc;
	sArgv = sArgv;
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamHandle, int nFlags, int nArgc, const char **sArgv)
{
	D(DLOG("pam_sm_setcred()\n"));
	return PAM_SUCCESS;

	// Unused Parameters
	pamHandle = pamHandle;
	nFlags = nFlags;
	nArgc = nArgc;
	sArgv = sArgv;
}

/*
 * Account management. We want to verify that the account exists 
 * before returning PAM_SUCCESS
 */
PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t *pamHandle, int nFlags, int nArgc, const char **sArgv)
{
	int			nRet = PAM_USER_UNKNOWN;
	const char	*sUsername;

	/* Get the username */
	nRet = pam_get_user(pamHandle, &sUsername, NULL);
	if ((nRet == PAM_SUCCESS) && (sUsername))
	{
		/* Verify the username */
		if (fuzzauth_UserExists(sUsername)) nRet = PAM_SUCCESS;
		else nRet = PAM_USER_UNKNOWN;
	}
	else
	{
		fuzzauth_Log(LOG_ERR, "pam_sm_acct_mgmt() - Couldn't get the username from PAM\n");
		nRet = PAM_SERVICE_ERR;
	}
	D(DLOG("pam_sm_acct_mgmt() - %s: %s\n", sUsername, nRet == PAM_SUCCESS ? "Success" : "Failure"));
	return nRet;

	// Unused Parameters
	nFlags = nFlags;
	nArgc = nArgc;
	sArgv = sArgv;
}

PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamHandle, int nFlags, int nArgc, const char **sArgv)
{
	int			nRet;
	const char	*sUsername;
	char			*sPasswordNew, *sPasswordOld;

	D(DLOG("pam_sm_chauthtok()\n"));
	/* Get name of a user */
	nRet = pam_get_user(pamHandle, &sUsername, "Username: ");
	if (nRet == PAM_SUCCESS)
	{
		/* Perform preliminary check and store requested password for updating later on */
		if (nFlags & PAM_PRELIM_CHECK)
		{
			struct pam_message	msg[3], *pmsg[3];
			struct pam_response	*resp;
			
			/* Converse to ensure we have the current password */
			nRet = PAM_Conversate(pamHandle);
			if (nRet == PAM_SUCCESS)
			{
				char *sChanging = NULL;
				
				asprintf(&sChanging, FA_PASSWORD_CHANGING, sUsername);
				if (sChanging)
				{
					/* Obtain and verify current password */
					pmsg[0] = &msg[0];
					msg[0].msg_style = PAM_TEXT_INFO;
					msg[0].msg = sChanging;
					pmsg[1] = &msg[1];
					msg[1].msg_style = PAM_PROMPT_ECHO_OFF;
					msg[1].msg = FA_PROMPT_PASSWORD_NEW;
					pmsg[2] = &msg[2];
					msg[2].msg_style = PAM_PROMPT_ECHO_OFF;
					msg[2].msg = FA_PROMPT_PASSWORD_CONFIRM;
					resp = NULL;
					nRet = PAM_Converse(pamHandle, 3, pmsg, &resp);
					if ((resp != NULL) && (nRet == PAM_SUCCESS))
					{
						/* Check password entered correctly */
						if (strcmp(resp[1].resp, resp[2].resp) != 0)
						{
							struct pam_response *resp2;
							msg[0].msg_style = PAM_ERROR_MSG;
							msg[0].msg = FA_PROMPT_PASSWORD_MISMATCH;
							PAM_Converse(pamHandle, 1, pmsg, &resp2);
							PAM_drop_reply(resp, 3);
							PAM_drop_reply(resp2, 1);
							nRet = PAM_AUTHTOK_RECOVER_ERR;
						}
						else
						{
							/* Store passwords */
							nRet = pam_set_item(pamHandle, PAM_OLDAUTHTOK, resp[1].resp);
							PAM_drop_reply(resp, 3);
						}
					}
					free(sChanging);
				}
				else nRet = PAM_AUTHTOK_RECOVER_ERR;
			}
		}
	}
	if (nFlags & PAM_UPDATE_AUTHTOK)
	{
		D(DLOG("pam_sm_chauthtok() - %s: Updating authentication token.\n", sUsername));
		nRet = pam_get_item(pamHandle, PAM_OLDAUTHTOK, (const void **)&sPasswordNew);
		if (nRet == PAM_SUCCESS)
		{
			nRet = pam_get_item(pamHandle, PAM_AUTHTOK, (const void **)&sPasswordOld);
			if (nRet == PAM_SUCCESS)
			{
				if (fuzzauth_UserExists(sUsername))
				{
					if (fuzzauth_ChangePassword((char *)sUsername, sPasswordOld, sPasswordNew)) nRet = PAM_SUCCESS;
					else nRet = PAM_PERM_DENIED;
				}
				else nRet = PAM_USER_UNKNOWN;
			}
		}
	}
	D(else DLOG("pam_sm_chauthtok() - %s: Not updating authentication token.\n", sUsername));
	D(DLOG("pam_sm_chauthtok() - %s: %s\n", sUsername, nRet == PAM_SUCCESS ? "Success" : "Failure"));
	return nRet;

	// Unused Parameters
	nArgc = nArgc;
	sArgv = sArgv;
}

PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamHandle, int nFlags, int nArgc, const char **sArgv)
{
	fuzzauth_Log(LOG_INFO, "PAM Session opened\n");
	return PAM_SUCCESS;	

	// Unused Parameters
	pamHandle = pamHandle;
	nFlags = nFlags;
	nArgc = nArgc;
	sArgv = sArgv;
}

PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamHandle, int nFlags, int nArgc, const char **sArgv)
{
	fuzzauth_Log(LOG_INFO, "PAM Session closed\n");
	return PAM_SUCCESS;

	// Unused Parameters
	pamHandle = pamHandle;
	nFlags = nFlags;
	nArgc = nArgc;
	sArgv = sArgv;
}

#ifdef PAM_STATIC
/* static module data */
struct pam_module _pam_fuzzauth_modstruct = {
	"pam_"FUZZAUTH_NAME_LOW,
	pam_sm_authenticate,
	pam_sm_setcred,
	pam_sm_acct_mgmt,
	pam_sm_open_session,
	pam_sm_close_session,
	pam_sm_chauthtok
};
#endif
